/*
 * distromatic - tool for RPM based repositories
 *
 * Copyright (C) 2004-2020 by Silvan Calarco <silvan.calarco@mambasoft.it>
 * Copyright (C) 2006 by Davide Madrisan <davide.madrisan@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of version 2 of the GNU General Public License as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, to the extent permitted by law; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

#include <assert.h>
#include <dirent.h>

#ifndef H_RPMLIB
# include <rpm/rpmlib.h>
#endif

#if RPM_VERSION >= 0x040100
#include <rpm/rpmts.h>
#else
#define rpmReadPackageFile(a,b,c,d) rpmReadPackageHeader(b,d,0,NULL,NULL)
#endif

#if HAVE_STRING_H
# if !STDC_HEADERS && HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif
#if HAVE_STRINGS_H
# include <strings.h>
#endif

#if STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# if HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#include "distromatic.h"
#include "config.h"
#include "changelog.h"
#include "functions.h"
#include "headerlist.h"
#include "rpmfunctions.h"
#include <errno.h>

#if !HAVE_MEMCPY
# define memcpy(d, s, n) bcopy ((s), (d), (n))
#endif

int rpmselector(const struct dirent *entry);
int sourcerpmselector(const struct dirent *entry);

const int bufsize = PATH_MAX;

struct warningList* addWarning(struct headerSourceList *pkg, char* text)
{
     struct warningList *currwarning, *prevwarning = NULL;

     if (!pkg) return NULL;

     currwarning = pkg->firstwarning;

     while (currwarning) {
          if (!strcmp(currwarning->text,text)) return currwarning;
          prevwarning = currwarning;
          currwarning = currwarning->next;
     }

     currwarning=malloc(sizeof(currwarning));
     if (!currwarning) return NULL;
     if (!pkg->firstwarning) pkg->firstwarning=currwarning;

     currwarning->text=strdup(text);
     if (!currwarning->text) return NULL;

     currwarning->next=NULL;
     if (prevwarning) prevwarning->next=currwarning;
     return pkg->firstwarning;
}

struct rebuildList* addRebuild(struct headerSourceList *pkg, struct headerSourceList* sourceheader, struct headerList* provider)
{
     struct rebuildList *currrebuild, *prevrebuild = NULL;

     if (!pkg) return NULL;

     currrebuild = pkg->firstrebuild;

     while (currrebuild) {
          if ((currrebuild->sourceheader == sourceheader) && (currrebuild->provider == provider)) return currrebuild;
          prevrebuild = currrebuild;
          currrebuild = currrebuild->next;
     }

     currrebuild=malloc(sizeof(struct rebuildList));
     if (!currrebuild) return NULL;
     if (!pkg->firstrebuild) pkg->firstrebuild=currrebuild;

     currrebuild->sourceheader=sourceheader;
     currrebuild->provider=provider;
     currrebuild->next=NULL;

     if (prevrebuild) prevrebuild->next=currrebuild;
     return pkg->firstrebuild;
}

struct headerList *
findPackageByName(struct headerList *list, char *name)
{
     struct headerList *currheaderlist;

     currheaderlist = list;

     while (currheaderlist) {
          if (strcmp(name, currheaderlist->name) == 0) {
               return currheaderlist;
          }
          currheaderlist = currheaderlist->next;
     }

     return NULL;
}

int getPackageNameFromFile(char *name)
{
     char *st1, *st2;

     st2 = strrchr(name, '-');
     if (st2) {
          *st2 = '\0';
          st1 = strrchr(name, '-');
          if (st1) {
               *st1 = '\0';
          } else {
               /* something wrong occurred: revert all */
               *st2 = '-';
	       return 1;
          }
     }
     return 0;
}

struct headerSourceList *
findSourcePackage(struct headerSourceList *list, char *name, char *version,
                  char *release, int altrepository)
{
     struct headerSourceList *currheaderlist = list;
     struct headerSourceList *oldheaderlist;

     while (currheaderlist) {
        if (!strcmp(name, currheaderlist->name)) {
           if (((!release) || !strcmp(release, currheaderlist->release)) &&
              ((!version) || !strcmp(version, currheaderlist->version)) &&
              ((altrepository < 0) || (altrepository == currheaderlist->altrepository)))
                 return currheaderlist;
           oldheaderlist = currheaderlist->old;
           while (oldheaderlist) {
              if (((!release) || !strcmp(release, oldheaderlist->release)) &&
                 ((!version) || !strcmp(version, oldheaderlist->version)) &&
                 ((altrepository < 0) || (altrepository == oldheaderlist->altrepository)))
                    return oldheaderlist;
              oldheaderlist = oldheaderlist->old;
           }
        }
        currheaderlist = currheaderlist->next;
     }

     return NULL;
}

int
getPackageInfoIntoHeaderList(Header h, struct headerList *hl)
{
     hl->name = headerGetStringEntry(h, RPMTAG_NAME);
     hl->epoch = headerGetUIntEntry(h, RPMTAG_EPOCH);
     if (hl->epoch == -1) hl->epoch = 0;
     hl->version = headerGetStringEntry(h, RPMTAG_VERSION);
     hl->release = headerGetStringEntry(h, RPMTAG_RELEASE);
     hl->summary = headerGetStringEntry(h, RPMTAG_SUMMARY);
     hl->arch = headerGetStringEntry(h, RPMTAG_ARCH);
     hl->description = headerGetStringEntry(h, RPMTAG_DESCRIPTION);
     hl->sourcename = headerGetStringEntry(h, RPMTAG_SOURCERPM);
     hl->group = headerGetStringEntry(h, RPMTAG_GROUP);
     hl->size = headerGetUIntEntry(h, RPMTAG_SIZE);

     return 0;
}

int
getPackageInfoIntoHeaderSourceList(Header h, struct headerSourceList *hl)
{
     char* packager;
     int count;

     hl->name = headerGetStringEntry(h, RPMTAG_NAME);
     hl->epoch = headerGetUIntEntry(h, RPMTAG_EPOCH);
     if (hl->epoch == -1) hl->epoch = 0;
     hl->version = headerGetStringEntry(h, RPMTAG_VERSION);
     hl->release = headerGetStringEntry(h, RPMTAG_RELEASE);
     hl->summary = headerGetStringEntry(h, RPMTAG_SUMMARY);
     hl->arch = headerGetStringEntry(h, RPMTAG_ARCH);
     hl->buildarchs = headerGetStringEntry(h, RPMTAG_BUILDARCHS);
     hl->excludearch = headerGetStringEntry(h, RPMTAG_EXCLUDEARCH);
     hl->description = headerGetStringEntry(h, RPMTAG_DESCRIPTION);
     packager = headerGetStringEntry(h, RPMTAG_PACKAGER);
     if (!packager) {
        logmsg(LOG_WARNING,"missing packager definition in package %s.",hl->name);
	    hl->packager = NULL;
     } else {
        hl->packager = getPackagerByName(packager,1);
        if (!hl->packager) {
           logmsg(LOG_WARNING,"cannot create '%s' packager for package %s.",packager,hl->name);
        free(packager);
	}
/*        if (! (hl->packager->role && PACKAGER_ROLE_MAINTAINER)) {
           logmsg(LOG_WARNING,"%s is an unmaintained package.",hl->name,hl->packager->name,hl->packager->role);
        }*/
     }
     hl->group = headerGetStringEntry(h, RPMTAG_GROUP);
     hl->license = headerGetStringEntry(h, RPMTAG_LICENSE);
     hl->url = headerGetStringEntry(h, RPMTAG_URL);
     if (!hl->url)
        logmsg(LOG_WARNING,"missing URL definition for package %s.",hl->name);
     hl->buildtime = headerGetUIntEntry(h, RPMTAG_BUILDTIME);

     hl->source = headerGetStringArrayEntry(h, RPMTAG_SOURCE, &count);
     hl->patch = headerGetStringArrayEntry(h, RPMTAG_PATCH, &count);
     hl->size = headerGetUIntEntry(h, RPMTAG_SIZE);

     return 0;
}

int
compareSourceHeaderBuildDate(const void *ptr1, const void *ptr2)
{
     return (int) ((*(struct headerSourceList **)ptr2)->buildtime -
                   (*(struct headerSourceList **)ptr1)->buildtime);
}

int
compareGroup(const void *ptr1, const void *ptr2)
{
     return (int) strcmp((*(struct headerSourceList **)ptr1)->group,
                   (*(struct headerSourceList **)ptr2)->group);
}

int
generateHeaderStats(struct configTag *ct, int arch)
{
     int c;
     struct headerList *currheaderlist;

     struct headerStats *stats = &ct->stats;
     struct headerList *headerlist = ct->headerlist[arch];
     int altrepository = ct->repository_level;

     stats->headercount[arch] = 0;
     stats->headersize[arch] = 0;
     currheaderlist = headerlist;
     while (currheaderlist) {
          if (currheaderlist->sourceheader
              && currheaderlist->altrepository == altrepository) {
               (stats->headercount[arch])++;
               stats->headersize[arch] += currheaderlist->size;
          }
          currheaderlist = currheaderlist->next;
     }

     stats->headerlistvec[arch] =
         malloc(sizeof(struct headerList *) * stats->headercount[arch]);
     if (!stats->headerlistvec) {
          return 1;
     }

     currheaderlist = headerlist;
     c = 0;
     while (currheaderlist) {
          if (currheaderlist->sourceheader
              && currheaderlist->altrepository ==
              altrepository) {
               stats->headerlistvec[arch][c] = currheaderlist;
               c++;
          }
          currheaderlist = currheaderlist->next;
     }
     return 0;
}

int
generateHeaderSourceStats(struct configTag *ct)
{
     int altrepository = ct->repository_level;
     struct headerSourceList *currheadersourcelist;
     struct headerSourceList *headersourcelist = ct->headersourcelist;
     struct headerStats *stats = &ct->stats;
     int c;
     struct changeLog *currchangelog;

     stats->headersourcecount = 0;
     stats->headersourcesize = 0;
     currheadersourcelist = headersourcelist;
     while (currheadersourcelist) {
          if (currheadersourcelist->altrepository == altrepository) {
               (stats->headersourcecount)++;
	       stats->headersourcesize += currheadersourcelist->size;
          }
          currheadersourcelist = currheadersourcelist->next;
     }

     stats->headersourcelistvec =
         malloc(sizeof(struct headerSourceList *) *
                stats->headersourcecount);
     if (!stats->headersourcelistvec) {
          return 1;
     }

     currheadersourcelist = headersourcelist;
     c = 0;
     ct->stats.changelog_total = 0;

     while (currheadersourcelist) {
          if (currheadersourcelist->altrepository == altrepository) {
               stats->headersourcelistvec[c] = currheadersourcelist;
               c++;
               if (currheadersourcelist->packager &&
	           (currheadersourcelist->packager->role & PACKAGER_ROLE_MAINTAINER)) {
                   currheadersourcelist->packager->packages_count++;
               }

               currchangelog = currheadersourcelist->changelog;


               while (currchangelog) {
                    ct->stats.changelog_total++;
                    if (currchangelog->pkg) {
                            currchangelog->pkg->changes_count++;
                    }
                    currchangelog = currchangelog->next;
               }

          }
          currheadersourcelist = currheadersourcelist->next;
     }

     return 0;
}

/*
 * this function should filter only RPM files
 * fixme: it currently just filters files from other entities */
int
rpmselector(const struct dirent *entry)
{
     char *ptr;

     ptr = strstr(entry->d_name, ".rpm");

     if (entry->d_type != DT_REG) return 0; /* skip if not a file */

     if (ptr == NULL) {
        logmsg(LOG_WARNING,"file %s has not rpm extension; skipping.",entry->d_name);
	return 0;
     }

     if (strstr(entry->d_name, ".src.rpm")) {
        logmsg(LOG_WARNING,"file %s looks like a source RPM; skipping.",entry->d_name);
	return 0;
    }

    return 1;
}

int
sourcerpmselector(const struct dirent *entry)
{
     char *ptr;

     ptr = strstr(entry->d_name, ".src.rpm");

     if (entry->d_type != DT_REG) return 0; /* skip if not a file */

     if (ptr == NULL) {
        logmsg(LOG_WARNING,"file %s has not src.rpm extension; skipping.",entry->d_name);
	return 0;
     }

     return 1;
}

long providedListCount[ARCHS_MAX];

void
cleanProvidedListIndex(struct configTag *ct, int arch)
{
   int i;

   for (i = 0; i < PROVIDEDLIST_IDX_SIZE; i++)
      ct->providedlist_idx[arch][i]=NULL;

   providedListCount[arch] = 0;
   ct->filetree[arch]=NULL;
}

void
createProvidedListIndex(struct providedList* *idx, int arch) {
   int i, step = providedListCount[arch]/PROVIDEDLIST_IDX_SIZE;
   int curridx = 1;
   struct providedList *currprovided = idx[0];

   if (providedListCount[arch] > PROVIDEDLIST_IDX_SIZE) {
      while (currprovided && curridx < PROVIDEDLIST_IDX_SIZE) {
         for (i = 0; i < step; i++)
            currprovided = currprovided->next;
         idx[curridx] = currprovided;
         curridx++;
      }
   }

}

static long providedListId = 0;

struct providedList*
findOrCreateProvidedListEntry(struct providedList* *idx,
                              char* findname, int create, int arch)
{
    struct providedList *newprovided, *currprovided, *prevprovided = NULL;
    int c = 0, i;
    int idxbot = 0, idxtop;

    if (idx[1]) {
       idxtop = PROVIDEDLIST_IDX_SIZE;
       while (idxtop-idxbot > 1) {
          i = (idxtop + idxbot) / 2;
          if (! idx[i]) idxtop = i;
          else {
             c = strcmp(idx[i]->name,findname);
             if (c < 0) idxbot = i;
             else if (c > 0) idxtop = i;
             else { idxbot = i-1; idxtop = i; }
          }
       }
//       printf("idxbot=%d idxtop=%d %s %s\n",idxbot,idxtop,idx[idxbot]->name,findname);

    }

    currprovided = idx[idxbot];

    while (currprovided &&
           ((c=strcmp(currprovided->name,findname)) < 0)) {
       prevprovided = currprovided;
       currprovided = currprovided->next;
    }

    if ((!currprovided) || (c > 0)) {
       if (!create) return NULL;
       providedListCount[arch]++;
       newprovided = malloc(sizeof(struct providedList));
       if (prevprovided) {
          newprovided->next = prevprovided->next;
          prevprovided->next = newprovided;
       } else {
          idx[0]=newprovided;
          newprovided->next = currprovided;
       }
       newprovided->provider = NULL;
       newprovided->numproviders = 0;
       newprovided->numbuildproviders = 0;
       newprovided->buildprovider = NULL;
       newprovided->name = strdup(findname);
       newprovided->id = ++providedListId;
       return newprovided;
    }
    return currprovided;
}

struct fileTree*
findOrCreateFileTreeBrother(struct fileTree* *first,char* findname, int arch)
{
    static long id[ARCHS_MAX] = { 0, 0, 0, 0, 0};
    struct fileTree *newdir, *currdir, *prevdir = NULL;
    int c = 0;

    currdir = *first;

    while (currdir &&
           ((c=strcmp(currdir->name,findname)) < 0)) {
       prevdir = currdir;
       currdir = currdir->next;
    }

    if ((!currdir) || (c > 0)) {
//       dirListCount++;
       newdir = malloc(sizeof(struct fileTree));
       newdir->firstchild = NULL;
       newdir->provider = NULL;
       newdir->numproviders = 0;
       newdir->parent = NULL;
       newdir->id = id[arch]++;

       if (prevdir) {
          newdir->next = prevdir->next;
          prevdir->next = newdir;
       } else {
          *first = newdir;
          newdir->next = currdir;
       }
       newdir->name = strdup(findname);
       return newdir;
    }

    return currdir;
}


struct fileTree*
findOrCreateFileTreeEntry(struct fileTree* *first,char* findname, int arch)
{
    struct fileTree *currdir,*prevdir=NULL;
    char *pstart,*pend;
    char f[PATH_MAX];
    int l;

    currdir = *first;
    pstart = findname+1; /* skip trailing slash */
    pend = findname+1;
    l = strlen(findname);

    while (pend-findname < l) {
       pend=strchr(pstart,'/');
       if (pend == 0) pend=findname+l; /* not final slash, it should be a file */

       strncpy(f,pstart,pend-pstart);
       f[pend-pstart]='\0';

       if (prevdir) {
          currdir = findOrCreateFileTreeBrother(&prevdir->firstchild,f,arch);
	  currdir->parent = prevdir;
       }
       else
          currdir = findOrCreateFileTreeBrother(first,f,arch);

       if (!*first)
          *first = currdir;
       prevdir = currdir;
       pstart=pend+1;
    }
//       fprintf(stderr,"%s (%d) firstchild:%d\n",currdir->name,prevdir,currdir->firstchild);

    return currdir;
}

struct fileUserList*
findOrCreateFileUserListEntry(struct fileUserList* *first,char* name, int arch)
{
   static long id[ARCHS_MAX] = { 0, 0, 0, 0, 0 };
   struct fileUserList *curr = *first, *prev = NULL, *newf = NULL;
   int i;

   while (curr && (i=strcmp(curr->name,name)) < 0) {
      prev = curr;
      curr = curr->next;
   }
   if (curr && (i == 0)) return curr;
   newf = malloc(sizeof(struct fileUserList));
   newf->name = strdup(name);
   newf->id = ++id[arch];
   if (prev) prev->next = newf; else *first = newf;
   if (curr) newf->next = curr; else newf->next = NULL;
   if (!*first) *first = newf;
   return newf;
}

struct fileGroupList*
findOrCreateFileGroupListEntry(struct fileGroupList* *first,char* name, int arch)
{
   static long id[ARCHS_MAX] = { 0, 0, 0, 0, 0 };
   struct fileGroupList *curr = *first, *prev = NULL, *newf = NULL;
   int i;

   while (curr && ((i=strcmp(curr->name,name)) < 0)) {
      prev = curr;
      curr = curr->next;
   }
   if (curr && (i == 0)) return curr;
   newf = malloc(sizeof(struct fileGroupList));
   newf->name = strdup(name);
   newf->id = ++id[arch];
   if (prev) prev->next = newf; else *first = newf;
   if (curr) newf->next = curr; else newf->next = NULL;
   if (!*first) *first = newf;
   return newf;
}

static long sourceid = 0;

int
addToSourceHeaderList(struct headerSourceList **headersourcelist, struct configTag *ct,
                      int mode, int altrepository)
{
     char *scanpath;
     struct headerSourceList *currheadersourcelist = NULL, *newheadersourcelist,
         *prevheadersourcelist;
     struct dirent **namelist;
     struct changeLog *changelog;
     Header h;
     char filepath[bufsize + 1];
     int n, j, arch, filenamescount, dirnamescount;
     int_16 *fileflags;
     char **basenames, **dirnames, **usernames, **groupnames;
     const char* errstr;
     int requirecount;
     char **requireversion, **requirename;
     uint_32 *requireflags, *dirindexes;
     char warning[PATH_MAX];
     rpmts ts = rpmtsCreate();

     if (altrepository == ct->repository_level) {
        scanpath = ct->repository_source_dir;
     } else if (altrepository == -1) {
        // non-incremental mode
        scanpath = ct->repository_source_dir;
        altrepository = 0;
     } else {
        scanpath = ct->repository[altrepository]->repository_source_dir;
     }
     assert(scanpath != NULL);
     memcpy(filepath, scanpath, strlen(scanpath) + 1);

     logmsg(LOG_DEBUG, "scanning source directory '%s' (%d)", scanpath, altrepository);
     n = scansdir(scanpath, &namelist, sourcerpmselector, scanrpmnamecmp);
     if (n < 0) {
          errstr = strerror(errno);
          logmsg(LOG_ERROR, "cannot scan directory '%s' (%s)", scanpath, errstr);
          return 1;
     }

     int cnt;
     for (cnt = 0; cnt < n; ++cnt) {
         /* check for duplicates */
         if ((cnt < n - 1) &&
	     (!rpmnamecmp(namelist[cnt]->d_name,namelist[cnt+1]->d_name,0))) {
             logmsg(LOG_WARNING,"skipping old SRPM package %s",namelist[cnt]->d_name);
	     cnt++;
	 }
         strcpy(&filepath[strlen(scanpath)], namelist[cnt]->d_name);

        logmsg(LOG_DEBUG, "getting header for %s", filepath);

    if (getHeader(&ts, filepath, &h)) {
	    errstr = strerror(errno);
            logmsg(LOG_WARNING,
   	               "%s: unable to read header (%s); skipping.",namelist[cnt]->d_name, errstr);
    } else {
           getPackageFiles(h, &dirindexes, &dirnames, &dirnamescount,
                           &basenames, &filenamescount,
                           &usernames, &groupnames, &fileflags);

           newheadersourcelist =
               malloc(sizeof(struct headerSourceList));
           if (newheadersourcelist == NULL) {
                return 1;
           }

           newheadersourcelist->next = NULL;
           for (arch = 0; arch < ARCHS_MAX; arch++) {
              newheadersourcelist->firstchild[arch] = NULL;
           }
           newheadersourcelist->updatingparent = NULL;
           newheadersourcelist->altrepository = altrepository;
           newheadersourcelist->firstwarning = NULL;
           newheadersourcelist->firstrebuild = NULL;
           newheadersourcelist->old = NULL;
           newheadersourcelist->id = ++sourceid;

           getPackageInfoIntoHeaderSourceList(h, newheadersourcelist);


           getPackageRequires(h, &requirename, &requireflags,
                              &requireversion, &requirecount);

           newheadersourcelist->dirindex =
               malloc(sizeof(int) * filenamescount);
           if (!newheadersourcelist->dirindex) {
                return 1;
           }
           memcpy(newheadersourcelist->dirindex, dirindexes,
               sizeof(int) * filenamescount);
           newheadersourcelist->dirname =
               (char **) dupnargv(dirnames, dirnamescount);
           newheadersourcelist->basename =
               (char **) dupnargv(basenames, filenamescount);
           newheadersourcelist->filenamecount = filenamescount;

	   newheadersourcelist->require = malloc(requirecount * sizeof(struct Require *));
	   for (j=0; j < requirecount; j++) {
	      newheadersourcelist->require[j] = malloc(sizeof(struct Require));
              newheadersourcelist->require[j]->name = requirename[j];
              newheadersourcelist->require[j]->flags = requireflags[j];
              newheadersourcelist->require[j]->version = requireversion[j];
	      newheadersourcelist->require[j]->resolved = NULL;
	   }
           newheadersourcelist->requirecount = requirecount;

           if (mode & GENHEADER_CHANGELOG) {
                changelog = getPackageChangelog(h, newheadersourcelist);
                while (changelog) {
                   if ((!changelog->pkg) || (!changelog->text)) {
                      snprintf(warning, PATH_MAX, "missing changelog name and/or text");
                      logmsg(LOG_WARNING,
                              "%s: missing changelog name and/or text",
                              newheadersourcelist->name);
                      addWarning(newheadersourcelist, warning);
                   }
                   changelog = changelog->next;
                }
           } else {
                newheadersourcelist->changelog = NULL;
           }

           /* if main repository (altrepository=0) packages are inserted sequentially
            * otherwise the list is scanned and packages are inserted alphabetically */
           if (!altrepository) {
                if (currheadersourcelist) {
                     currheadersourcelist->next = newheadersourcelist;
                }
                currheadersourcelist = newheadersourcelist;
           } else { /* altrepository */
                currheadersourcelist = *headersourcelist;
                prevheadersourcelist = NULL;

                while ((currheadersourcelist) &&
                       (strcasecmp(currheadersourcelist->name,newheadersourcelist->name) < 0)) {
                     prevheadersourcelist = currheadersourcelist;
                     currheadersourcelist = currheadersourcelist->next;
                 }

                 if ((currheadersourcelist) &&
                    (!strcmp(currheadersourcelist->name,newheadersourcelist->name))) {
 		    /* the package is both in main and alternate repositories.
		       override main repository package */
                    if ((currheadersourcelist->epoch > newheadersourcelist->epoch) ||
                        ((currheadersourcelist->epoch == newheadersourcelist->epoch) &&
		         (rpmvercmp
                         (currheadersourcelist->version,
                          newheadersourcelist->version) > 0)) ||
                        ((currheadersourcelist->epoch == newheadersourcelist->epoch) &&
		         (rpmvercmp
                         (currheadersourcelist->version,
                          newheadersourcelist->version) == 0) &&
                         (rpmvercmp
                          (currheadersourcelist->release,
                           newheadersourcelist->release) >= 0))
			 ) {
			     if (ct->repository_level == altrepository) {
                                snprintf(warning, PATH_MAX, "is older or same release of upstream package in %s (%ld:%s-%s <= %ld:%s-%s)",
                                   ct->repository[currheadersourcelist->altrepository]->tag,
                                   newheadersourcelist->epoch,
                                   newheadersourcelist->version,
                                   newheadersourcelist->release,
                                   currheadersourcelist->epoch,
                                   currheadersourcelist->version,
                                   currheadersourcelist->release);
                                logmsg(LOG_WARNING,
                                   "%s(source,%s): %s",
                                   newheadersourcelist->name,
                                   ct->repository[newheadersourcelist->altrepository]->tag,
                                   warning);
                                addWarning(newheadersourcelist, warning);
                            }
                       }
		       newheadersourcelist->next = currheadersourcelist->next;
		       /* overriden package is not deallocated but referenced with currheadersourcelist->old */
		       newheadersourcelist->old=currheadersourcelist;
		       newheadersourcelist->old->updatingparent=newheadersourcelist;
                    } else {
		       /* add the new package to the list */
                       newheadersourcelist->next = currheadersourcelist;
		    }
                    if (prevheadersourcelist) {
                       prevheadersourcelist->next = newheadersourcelist;
                    } else {
                       *headersourcelist = newheadersourcelist;
                    }
               }

               if (!*headersourcelist) {
                    /* set first pointer of the list */
                    *headersourcelist = newheadersourcelist;
           } /* altrepository */
           (void) headerFree(h);
       } // if getHeader()
       free(namelist[cnt]);
     } // for

     free(namelist);
     rpmtsFree(ts);
     return 0;
}

int
generateSourceHeaderList(struct configTag *ct, int mode, int incremental)
{
     struct headerSourceList **headersourcelist = &(ct->headersourcelist);
     int i;

     if (incremental) {
        for (i = 0; i <= ct->repository_level; i++) {
           if (ct->repository[i]) {
              if (addToSourceHeaderList(headersourcelist, ct, mode, i)) {
                 logmsg(LOG_ERROR,"Error scanning alternate SRPMs repository");
                 return 1;
              }
           } else {
              break;
           }
        }
     } else {
        if (addToSourceHeaderList(headersourcelist, ct, mode, -1)) {
             logmsg(LOG_ERROR,"Error scanning alternate SRPMs repository");
             return 1;
        }
     }

     return 0;
}

/*
 * Reads all headers and generates a structure based on struct headerList */
int
addToHeaderList(struct configTag *ct,
                const char *scanpath, /*int mode,*/
                const char *scantag,
                const char *altscanpath[ALT_REPS_MAX],
                const char *altscantag[ALT_REPS_MAX],
                int arch)
{

/*#headerList **headerlist,
#                struct headerSourceList *headersourcelist,
#                struct providedList **firstprovided,
*/

     struct headerList *currheaderlist, *newheaderlist, *currbrother;
     struct providedList *provided;
     struct headerList ** newprovider;
     struct dirent **namelist, **altnamelist[ALT_REPS_MAX];
     Header h;
     char filepath[bufsize + 1], currname[bufsize + 1];
     char *filename=NULL;
     long i ,j , k, n, altn[ALT_REPS_MAX];
     int altrepository=0, obsoletecount, providecount, requirecount,
         filenamecount, dirnamecount;
     int_16 *fileflags;
     char **obsoletename, **obsoleteversion,
          **providename, **provideversion,
          **requirename, **requireversion,
         **basename, **dirname, **newversion,
         **fileusername, **filegroupname;
     const char* errstr;
     uint_32 *dirindex, *requireflags, *obsoleteflags, *provideflags;
#if RPM_VERSION >= 0x050000
     rpmts ts = rpmtsCreate();
#else
     rpmts ts = NULL;
#endif

     currheaderlist = ct->headerlist[arch];

     n = scansdir(scanpath, &namelist, rpmselector, scanrpmnamecmp);

     if (n < 0) {
          errstr = strerror(errno);
          logmsg(LOG_ERROR, "%s: cannot scan directory '%s' (%s)", scantag, scanpath, errstr);
          return 1;
     }
     logmsg(LOG_DEBUG, "%s scan directory is '%s'", scantag, scanpath);
     for (i = 0; i < ALT_REPS_MAX; i++) {
        if (altscanpath[i]) {
           altn[i] = scansdir(altscanpath[i], &altnamelist[i], rpmselector, scanrpmnamecmp);
           logmsg(LOG_DEBUG, "%s scan directory is '%s'", altscantag[i], altscanpath[i]);
           if (altn[i] < 0) {
              errstr = strerror(errno);
              logmsg(LOG_ERROR, "%s: cannot scan directory '%s' (%s)", altscantag[i], altscanpath[i], errstr);
              return 1;
           }
        } else { break; }
     }
     if (i < ALT_REPS_MAX) {
        altnamelist[i]=NULL;
     }

     int cnt = 0, altcnt[ALT_REPS_MAX];
     for (i = 0; i < ALT_REPS_MAX; i++) {
        altcnt[i]=0;
     }

     while (1) {
/*        if (cnt>=n) { break; }
        for (i = 0; i < ALT_REPS_MAX; i++) {
	   if altcnt[i] >= altn[i]
	}*/
        /* check for duplicates in main repository */

        if ((cnt < n - 1) &&
	    (!rpmnamecmp(namelist[cnt]->d_name,namelist[cnt+1]->d_name,0))) {
            logmsg(LOG_WARNING, "%s: duplicated RPM package in %s repository (skipped)",namelist[cnt]->d_name, scantag);
	    cnt++;
	}
	/* find next package in all repositories */
       if (cnt < n) {
           strncpy(currname, namelist[cnt]->d_name, bufsize);
	} else {
           currname[0] = '\0';
	}
        for (i = 0; i < ALT_REPS_MAX; i++) {
	   /* scan alternate repositories to find any package which comes first according to name sorting */
	   if (altscanpath[i] && (altcnt[i] < altn[i]) &&
              ((rpmnamecmp(currname, altnamelist[i][altcnt[i]]->d_name,0) >= 0) || (currname[0] == '\0'))) {
              strncpy(currname, altnamelist[i][altcnt[i]]->d_name,bufsize);
           }
	}

	if (cnt < n && (rpmnamecmp(currname, namelist[cnt]->d_name, 0) == 0)) {
	   /* selected package exists in base repository */
           memcpy(filepath, scanpath, strlen(scanpath) + 1);
	   filename=&filepath[strlen(scanpath)];
           strncpy(filename, namelist[cnt]->d_name,bufsize-strlen(scanpath));
	   altrepository=0;
	   cnt++;
	} else {
           for (i = 0; i < ALT_REPS_MAX; i++) {
	      if (altscanpath[i] && (altcnt[i] < altn[i]) &&
	         (strncmp(currname, altnamelist[i][altcnt[i]]->d_name, bufsize) == 0)) {
	         /* selected package exists in this repository */
                 memcpy(filepath, altscanpath[i], strlen(altscanpath[i]) + 1);
	         filename=&filepath[strlen(altscanpath[i])];
                 strncpy(filename, (altnamelist[i])[altcnt[i]]->d_name, bufsize-strlen(altscanpath[i]));
                 altrepository=i+1;
	         altcnt[i]++;
                 /* check for duplicates in alternate repository */
                 if ((altcnt[i] < altn[i] - 1) &&
    	             (!rpmnamecmp(altnamelist[i][altcnt[i]]->d_name,altnamelist[i][altcnt[i]+1]->d_name,0))) {
    	             if (i == ct->repository_level-1) {
                        logmsg(LOG_WARNING, "%s: duplicated RPM package (skipped)",
                           altnamelist[i][altcnt[i]]->d_name, altscantag[i]);
                     }
	         }
	         break;
              }
           }
        }

	/* break while loop when no more packages are found */
	if (currname[0] == '\0') { break; }
	/* process package */
        logmsg(LOG_DEBUG, "getting header for %s", filepath);

        if (getHeader(&ts, filepath, &h)) {
            errstr = strerror(errno);
            logmsg(LOG_WARNING,
   	               "%s: unable to read header (%s); skipping.",filename, errstr);
        } else {

           getPackageObsoletes(h, &obsoletename, &obsoleteflags,
                              &obsoleteversion, &obsoletecount);
           getPackageProvides(h, &providename, &provideflags,
                              &provideversion, &providecount);
           getPackageRequires(h, &requirename, &requireflags,
                              &requireversion, &requirecount);
           getPackageFiles(h, &dirindex, &dirname, &dirnamecount,
                           &basename, &filenamecount,
                           &fileusername, &filegroupname, &fileflags);

           newheaderlist = malloc(sizeof(struct headerList));
           if (newheaderlist == NULL) {
	        logmsg(LOG_ERROR,"header in file %s is corrupted; aborting.\n",filepath);
                return 1;
           }
           memset(newheaderlist, 0, sizeof(struct headerList));
           getPackageInfoIntoHeaderList(h, newheaderlist);

	   if (!newheaderlist->sourcename) {
	      logmsg(LOG_WARNING,
                      "sourcename undefined in RPM file %s, it might be a source RPM with wrong filename; skipping package.",
		      filepath);
	      newheaderlist->sourceheader = NULL;
              (void) headerFree(h);
              free(newheaderlist);
	      continue;
	   } else {
	      if (getPackageNameFromFile(newheaderlist->sourcename)) {
	         logmsg(LOG_WARNING,
		      "file %s doesn't have a standard format; skipping package.");
		 newheaderlist->sourceheader = NULL;
                 (void) headerFree(h);
		 free(newheaderlist);
		 continue;
	      } else {
                 newheaderlist->sourceheader =
                    findSourcePackage(ct->headersourcelist,
                                    newheaderlist->sourcename,
                                    newheaderlist->version,
                                    newheaderlist->release,
                                    altrepository);
	      }
           }
           if (newheaderlist->sourceheader) {
                if (!newheaderlist->sourceheader->firstchild[arch]) {
                     newheaderlist->sourceheader->firstchild[arch] = newheaderlist;
                } else {
                     currbrother =
                         newheaderlist->sourceheader->firstchild[arch];
                     while (currbrother->nextbrother) {
                          currbrother = currbrother->nextbrother;
                     }
                     currbrother->nextbrother = newheaderlist;
                }

                newheaderlist->recursed = 0;

                if (currheaderlist) {
                      currheaderlist->next = newheaderlist;
                }
                currheaderlist = newheaderlist;

		if (!ct->headerlist[arch]) {
                   /* set first pointer of the list */
                   ct->headerlist[arch] = newheaderlist;
                }

	   } else if (newheaderlist->sourcename) /* missing source header */ {

                if ((!newheaderlist->sourceheader) ||
		    (newheaderlist->sourceheader->altrepository) <= altrepository)
		{
		   logmsg(LOG_WARNING,"%s(%s,%s): missing SRPM %s-%s-%s; skipping.",
		   newheaderlist->name,
		   ct->arch[arch],
		   ct->repository[altrepository]->tag,
                   newheaderlist->sourcename,
                   newheaderlist->version, newheaderlist->release);
                   (void) headerFree(h);
                   free(newheaderlist);
                   continue;
                }
	    } // if newheaderlist->sourceheader

            newheaderlist->provided = malloc(sizeof(struct providedList*)*providecount);
            for (i=0; i < providecount; i++) {
	       provided = findOrCreateProvidedListEntry((struct providedList **) &(ct->providedlist_idx[arch]),
	                                                providename[i],1,arch);
	       newheaderlist->provided[i]=provided;
	       if (provided && (provided->numproviders == 0)) {
	          provided->flags=provideflags[i];
	          provided->numproviders++;
	          provided->provider=malloc(sizeof(struct headerList*));
	          provided->provider[0]=newheaderlist;
	          provided->version=malloc(sizeof(char *));
		  provided->version[0]=strdup(provideversion[i]);
	       } else if (provided && (provided->numproviders > 0)) {
		  provided->numproviders++;
		  newprovider=malloc(sizeof(struct headerList*)*provided->numproviders);
                  for (j = 0; j < provided->numproviders-1; j++) {
		     newprovider[j]=provided->provider[j];
		  }
		  newprovider[provided->numproviders-1] = newheaderlist;
		  free(provided->provider);
		  provided->provider=newprovider;

		  newversion=malloc(sizeof(char *)*provided->numproviders);
                  for (j = 0; j < provided->numproviders-1; j++) {
		     newversion[j]=provided->version[j];
		  }
		  newversion[provided->numproviders-1] = strdup(provideversion[i]);
		  free(provided->version);
		  provided->version=newversion;

	       } else {
	          fprintf(stderr,"%s has %d providers but is ignored\n",provided->name,provided->numproviders);
	       }
	    }

           newheaderlist->providename = providename;
           newheaderlist->provideflags = provideflags;
           newheaderlist->provideversion = provideversion;
           newheaderlist->providecount = providecount;

	   newheaderlist->obsoletename = obsoletename;
           newheaderlist->obsoleteflags = obsoleteflags;
           newheaderlist->obsoleteversion = obsoleteversion;
           newheaderlist->obsoletecount = obsoletecount;

           newheaderlist->altrepository = altrepository;

  	   newheaderlist->require = malloc(requirecount * sizeof(struct Require *));
           for (j=0; j < requirecount; j++) {
              newheaderlist->require[j] = malloc(sizeof(struct Require));
              newheaderlist->require[j]->name = requirename[j];
              newheaderlist->require[j]->flags = requireflags[j];
	      newheaderlist->require[j]->version = requireversion[j];
	      newheaderlist->require[j]->resolved = NULL;
	   }
           newheaderlist->requirecount = requirecount;

	   newheaderlist->file =
	       malloc(sizeof(struct fileTree*) * filenamecount);
	   newheaderlist->fileflags =
	       malloc(sizeof(int_16 *) * filenamecount);
	   newheaderlist->fileuser =
	       malloc(sizeof(struct fileUserList *) * filenamecount);
	   newheaderlist->filegroup =
	       malloc(sizeof(char *) * filenamecount);
	   for (j=0; j<filenamecount; j++) {
               snprintf(filename,bufsize,"%s%s",dirname[dirindex[j]],basename[j]);
               newheaderlist->file[j] = findOrCreateFileTreeEntry(&ct->filetree[arch], filename, arch);
               newheaderlist->fileflags[j] = fileflags[j];
               newheaderlist->fileuser[j] = findOrCreateFileUserListEntry(&ct->fileuserlist[arch], fileusername[j], arch);
               free(fileusername[j]);
               newheaderlist->filegroup[j] = findOrCreateFileGroupListEntry(&ct->filegrouplist[arch], filegroupname[j], arch);
               free(filegroupname[j]);
               if (newheaderlist->file[j]->numproviders == 0) {
                     newheaderlist->file[j]->numproviders++;
                     newheaderlist->file[j]->provider=malloc(sizeof(struct headerList*));
                     newheaderlist->file[j]->provider[0]=newheaderlist;
               } else {
		     /* check if the package is already providing it... */
		     for (k = 0; k < newheaderlist->file[j]->numproviders; k++) {
		        if (newheaderlist->file[j]->provider[k] == newheaderlist)
			   break;
		     }

		     /* if not add to the list of providers */
		     if (k == newheaderlist->file[j]->numproviders) {
		        newheaderlist->file[j]->numproviders++;
			newprovider=malloc(sizeof(struct headerList*)*newheaderlist->file[j]->numproviders);
                        for (k = 0; k < newheaderlist->file[j]->numproviders-1; k++) {
			   newprovider[k]=newheaderlist->file[j]->provider[k];
			}
			newprovider[newheaderlist->file[j]->numproviders-1] = newheaderlist;
			free(newheaderlist->file[j]->provider);
			newheaderlist->file[j]->provider=newprovider;
		     }
               }
	   }
           newheaderlist->filenamecount = filenamecount;
           (void) headerFree(h);
	} // if getHeader()
        if (altrepository > 0) {
	    free(altnamelist[altrepository-1][altcnt[altrepository-1]-1]); }
	else free(namelist[cnt-1]);
     } // main while
#if RPM_VERSION >= 0x050000
     rpmtsFree(ts);
#endif
     free(namelist);
     createProvidedListIndex((struct providedList **)&(ct->providedlist_idx[arch]), arch);
     return 0;
}

int
generateHeaderList(struct configTag* ct, int arch, int incremental)
{
     char scanpath[bufsize + 1], *altscanpath[ALT_REPS_MAX];
     char *scantag, *altscantag[ALT_REPS_MAX];
     int i;

     for (i = 0; i < ALT_REPS_MAX; i++)
        altscanpath[i] = NULL;

     cleanProvidedListIndex(ct,arch);

     if (incremental) {
        if (ct->repository[0]) {
           snprintf(scanpath,bufsize,"%sRPMS.%s/",ct->repository[0]->repository_dir,ct->arch[arch]);
           scantag = ct->repository[0]->tag;
           for (i = 1; i <= ct->repository_level; i++) {
              if (ct->repository[i]) {
                 altscanpath[i-1] = malloc(bufsize);
                 snprintf(altscanpath[i-1], bufsize, "%sRPMS.%s/",ct->repository[i]->repository_dir,ct->arch[arch]);
                 altscantag[i-1] = ct->repository[i]->tag;
               } else {
                 logmsg(LOG_ERROR,"repository[i] not defined; aborting.");
                 return 1;
              }
           }
        } else {
           logmsg(LOG_ERROR,"repository[0] is not defined; aborting.");
           return 1;
        }
     } else {
        snprintf(scanpath,bufsize,"%sRPMS.%s/",ct->repository_dir,ct->arch[arch]);
        scantag = ct->tag;
     }
     if (addToHeaderList(ct, scanpath, scantag, (const char**)altscanpath, (const char**)altscantag, arch)) {
          logmsg(LOG_ERROR,"Error scanning RPMs repository");
          return 1;
     }

     return 0;
}
